/*
 * fe5680 - Tames your FE-5680A Rubidium Atomic Frequency Standard
 * Copyright (C) 2014  Tom Wimmenhove <tom.wimmenhove@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <syslog.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/select.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <getopt.h>
#include <math.h>
#include <signal.h>
#include <sys/time.h>
#include "fe5680.h"
#include "config.h"

#ifdef DEBUG
int debug = 0;

void debug_print(char *s, unsigned char *data, int n)
{
	int i;

	if (!debug)
		return;

	printf("%s: ", s);
	for (i=0;i<n;i++)
		printf("%02X ", data[i]);
	printf("\n");
}
#endif

int serread(int fd, unsigned char *buf, int len)
{
	int ret = 0;
	int rd = 0;
	fd_set rfds;
	struct timeval tv;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);

	tv.tv_sec = 1;
	tv.tv_usec = 0;
	while (rd < len) {
		if (select(fd+1, &rfds, NULL, NULL, &tv) == -1)
			return -1;

		if (!FD_ISSET(fd, &rfds))
			return -1;

		ret = read(fd, buf + rd, len - rd);

		if (ret == 0)
			return 0;

		rd += ret;
	}

	if (ret <= 0)
		return -1;

	return len;
}

int fe5680a_send_packet(int fd, unsigned char command,
	unsigned char * payload, unsigned short len)
{
	unsigned char pkt[4 + len + 1];
	unsigned char chsum = 0;
	unsigned char b;
	int pktlen;
	int i;

	pktlen = 4 + len;

	if (len)
	{
		for (i=0;i<len;i++)
		{
			b = payload[i];
	
			chsum ^= b;
			pkt[4 + i] = b;
		}
		pkt[4 + i] = chsum;

		// room for the checksum
		pktlen++;
	}

	/* Command header */
	pkt[0] = command;
	pkt[1] = pktlen & 0xff;
	pkt[2] = (pktlen >> 8) & 0xff;
	pkt[3] = pkt[0] ^ pkt[1] ^ pkt[2];

#ifdef DEBUG
	debug_print("Command", pkt, pktlen);
#endif

	return write(fd, pkt, pktlen);
}

int fe5680a_read_response(int fd, unsigned char * response)
{
	unsigned char chsum;
	int i, len;

	/* Read response header */
	serread(fd, response, 4);
	/* Checksum */
	if (response[0] ^ response[1] ^ response[2] ^ response[3])
		return -1;

	len = (response[1] | (response[2] << 8));

	/* If there's a payload comming */
	if (len > 4)
	{
		chsum = 0;

		/* Read payload */
		if (serread(fd, response + 4, len - 4) == -1)
			return -1;

		/* Checksum */
		for (i=4;i<len;i++)
			chsum ^= response[i];

		if (chsum)
			return -1;
	}

#ifdef DEBUG
	debug_print("Response", response, len);
#endif

	return len;
}

int fe5680a_set_freq_offset_save(int fd, int offset)
{
	unsigned char pkt[4];

	pkt[0] = (offset >> 24) & 0xff;
	pkt[1] = (offset >> 16) & 0xff;
	pkt[2] = (offset >>  8) & 0xff;
	pkt[3] = (offset >>  0) & 0xff;

	return fe5680a_send_packet(fd, 0x2c, pkt, 4);
}


int fe5680a_set_freq_offset(int fd, int offset)
{
	unsigned char pkt[4];

	pkt[0] = (offset >> 24) & 0xff;
	pkt[1] = (offset >> 16) & 0xff;
	pkt[2] = (offset >>  8) & 0xff;
	pkt[3] = (offset >>  0) & 0xff;

	return fe5680a_send_packet(fd, 0x2e, pkt, 4);
}

int fe5680a_get_freq_offset(int fd, int *offset)
{
	unsigned char response[256];
	int ret;

	ret = fe5680a_send_packet(fd, 0x2d, NULL, 0);
	if (ret == -1)
		return -1;

	ret = fe5680a_read_response(fd, response);
	if (ret == -1)
		return -1;

	*offset = (response[4] << 24) | (response[5] << 16) |
		(response[6] << 8) | response[7];

	return 0;
}

/*
 * fe5680a_get_serial(), fe5680a_get_adc() and fe5680a_get_dds() use undocumented commands.
 * See http://ko4bb.com/dokuwiki/doku.php?id=precision_timing:fe5680a_faq
 *     https://www.febo.com/pipermail/time-nuts/2012-January/062866.html
 *     https://www.febo.com/pipermail/time-nuts/2012-January/062861.html
 */
int fe5680a_get_serial(int fd, char *serial)
{
	unsigned char response[256];
	int i, ret, len;

	ret = fe5680a_send_packet(fd, 0x61, NULL, 0);
	if (ret == -1)
		return -1;

	len = fe5680a_read_response(fd, response);
	if (len == -1)
		return -1;

	ret = 0;
	for (i=4;i<len - 1;i++)
		ret += sprintf(serial + ret, "%02x", response[i]);

	return 0;
}

int fe5680a_get_adc(int fd, unsigned short counts[4])
{
	unsigned char response[256];
	int i, ret;

	ret = fe5680a_send_packet(fd, 0x5a, NULL, 0);
	if (ret == -1)
		return -1;

	ret = fe5680a_read_response(fd, response);
	if (ret == -1)
		return -1;

	for (i=0;i<4;i++)
		counts[i] = response[4 + 2 * i] |
			   (response[4 + 2 * i + 1] << 8);

	return 0;
}

int fe5680a_get_dds(int fd, long int *dds)
{
	unsigned char response[256];
	int ret;

	ret = fe5680a_send_packet(fd, 0x22, NULL, 0);
	if (ret == -1)
		return -1;

	ret = fe5680a_read_response(fd, response);
	if (ret == -1)
		return -1;

	*dds =  ((long int) response[ 4] << 56) |
		((long int) response[ 5] << 48) |
		((long int) response[ 6] << 40) |
		((long int) response[ 7] << 32) |
		((long int) response[ 8] << 24) |
		((long int) response[ 9] << 16) |
		((long int) response[10] <<  8) |
		 (long int) response[11];

	return 0;
}

void commerror(int die)
{
	fprintf(stderr, "Couldn't read response\n");
	if (die)
		exit(1);
}

volatile int tick = 0;

void timer_handler (int signum)
{
	tick = 1;
}


int main (int argc, char **argv)
{
	struct termios options;
	char *dev;
	int fd;
	double offs;
	unsigned short counts[4];
	int i, n;
	int offset, roffset;
	long int dds;
	int dosave = 0;
	int doread = 0;
	int doset = 0;
	int dosiunits = 0;
	int doserial = 0;
	int doadc = 0;
	int dodds = 0;
	int c;
	char s[256];
	char *adcnames[4] = { "???", "Side pot", "Temp?", "VCXO?" };
	struct sigaction sa;
	struct itimerval timer;

	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{ "offset", 1, 0, 'o' },
			{ "save", 0, 0, 's' },
			{ "read", 0, 0, 'r' },
			{ "hz", 0, 0, 'u' },
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'v'},
// and the undocumented ones
			{ "serial", 0, 0, 'S' },
			{ "adc", 0, 0, 'a' },
			{ "dds", 0, 0, 'y' },
#ifdef DEBUG
			{ "debug", 0, 0, 'd' },
#endif
			{ 0, 0, 0, 0}
		};
		c = getopt_long(argc, argv, "ao:sruhSdyv",
				long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'a':
				doadc = 1;
				break;
			case 'o':
				offs = atof(optarg);
				offset = lround(offs);
				doset = 1;
				break;
			case 'u':
				dosiunits = 1;
				break;
			case 's':
				dosave = 1;
				break;
			case 'S':
				doserial = 1;
				break;
			case 'r':
				doread = 1;
				break;
			case 'y':
				dodds = 1;
				break;
			case 'h':
				printf("Usage: %s <options> <serial port>\n", argv[0]);
				printf("	-o, --offset <offset> : set offset (range -2147483648 to 2147483647)\n");
				printf("	-s, --save            : save offset to FE5680A's EEPROM\n");
				printf("	-r, --read            : read offset\n");
				printf("	-u, --hz              : use units in Hz instead of raw counts\n");
#ifdef DEBUG
				printf("	-d, --debug           : print debug shizzle\n");
#endif
				printf("	-v, --version         : print version number and exit\n");
				printf("	-h, --help            : This\n");
				printf("         --- Undocumented FE5680A commands: READ MANUAL, USA AT OWN RISK!---\n");
				printf("	-S, --serial          : Read serial number from device\n");
				printf("	-a, --adc             : Log ADC counts every second\n");
				printf("	-y, --dds             : Read the current DDS Value\n");
				exit(0);
#ifdef DEBUG
			case 'd':
				debug = 1;
				break;
#endif

			case 'v':
				printf("fe5680 Version %d.%d by Tom Wimmenhove <tom.wimmenhove@gmail.com>\n", VERSION_MAJOR, VERSION_MINOR);
				exit(0);

			default:
				printf("Try %s --help\n", argv[0]);
				exit(1);
		}
	}

	if (optind >= argc)
	{
		fprintf(stderr, "Give serial port name as parameter,"
				" try %s --help\n", argv[0]);
		exit(1);
	}

	if (dosiunits)
	{
		offset = FREQ_TO_OFFSET(offs);
	}

	dev = argv[optind];

	fd = open (dev, O_RDWR);


	if (fd == -1) {
		perror(dev);
		return -1;
	}

	memset(&options, 0, sizeof(options));
	options.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
			| INLCR | IGNCR | ICRNL | IXON);
	options.c_oflag &= ~OPOST;
	options.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	options.c_cflag &= ~(CSIZE | PARENB);
	options.c_cflag |= CS8;

	cfsetispeed(&options, B9600);
	cfsetospeed(&options, B9600);

	options.c_cflag |= CLOCAL | CREAD | CRTSCTS;

	tcsetattr(fd, TCSANOW, &options);

	tcflush(fd, TCIOFLUSH);

	if (doread)
	{
		if (fe5680a_get_freq_offset(fd, &roffset) == -1)
			commerror(1);

		printf("Frequency offset: %d (%fHz)\n", roffset,
				OFFSET_TO_FREQ(roffset));
	}

	/* read offset from clock in case of -save without a given offset */
	if (dosave && (!doset))
	{
		if (fe5680a_get_freq_offset(fd, &offset) == -1)
			commerror(1);

		doset = 1;
	}

	if (doserial)
	{
		if (fe5680a_get_serial(fd, s) == -1)
			commerror(1);

		printf("Serial #: %s\n", s);
	}


	if (dodds)
	{
		if (fe5680a_get_dds(fd, &dds) == -1)
			commerror(1);

		printf("DDS Value: %ld (0x%lx)\n", dds, dds);
	}



	if (doset)
	{
		if (dosave)
		{
			printf("Saving frequency offset: %d (%fHz) to EEPROM\n",
				       	offset, OFFSET_TO_FREQ(offset));
			fe5680a_set_freq_offset_save(fd, offset);
		}
		else
		{
			printf("Setting frequency offset: %d (%fHz)\n", offset,
					OFFSET_TO_FREQ(offset));
			fe5680a_set_freq_offset(fd, offset);
		}
	}


	if (doadc)
	{
		memset (&sa, 0, sizeof (sa));
		sa.sa_handler = &timer_handler;
		sigaction (SIGALRM, &sa, NULL);

		timer.it_value.tv_sec = 1;
		timer.it_value.tv_usec = 0;
		timer.it_interval.tv_sec = 1;
		timer.it_interval.tv_usec = 0;

		setitimer (ITIMER_REAL, &timer, NULL);

		for (i=0;i<4;i++)
			printf("# Ch%d: %s\n",  i+1, adcnames[i]);

		n = 0;
		printf("#  Time (s) | Ch1 | Ch2 | Ch3 | Ch4\n");
		while (1)
		{
			if (fe5680a_get_adc(fd, counts) == -1)
			{
				commerror(0);

				n++;
				continue;
			}

			printf("%10d ", n);
			for (i=0;i<4;i++)
				printf("%5d ",counts[i]);
			printf("\n");
			fflush(stdout);

			n++;

			// Wait for next second
			do {
				pause();
			} while (!tick);
			tick = 0;			
		}
 
	}


	return 0;
}


